//
//  G15Device.m
//  G15
//
//  Created by Casey Fleser on 9/12/07.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

#import "G15Device.h"
#import "G15Manager.h"
#import "G15Display.h"
#import <IOKit/IOKitLib.h>
#import <IOKit/IOCFPlugIn.h>

void G15HIDReportCallbackFunction(
	void *	 		inTarget,
	IOReturn 		inResult,
	void * 			inRefcon,
	void * 			inSender,
	uint32_t		 	inBufferSize)
{
	if (inResult == kIOReturnSuccess) {
		G15Device		*device = (G15Device *)inTarget;

		[device processReadData: inBufferSize];
	}
}

@implementation G15Device

+ (UInt32) locationID: (io_service_t) inServiceID
{
	NSDictionary			*properties;
	NSNumber				*location;
	
	properties = [G15Device deviceProperties: inServiceID];
	location = [properties valueForKey: (NSString *)CFSTR(kIOHIDLocationIDKey)];

	return location != nil ? [location unsignedLongValue] : 0;
}

+ (NSDictionary *) deviceProperties: (io_service_t) inServiceID
{
	NSMutableDictionary		*properties = nil;
	IOReturn				result;

	result = IORegistryEntryCreateCFProperties(inServiceID, (CFMutableDictionaryRef *)&properties, kCFAllocatorDefault, kNilOptions);
		
	return [properties autorelease];
}

- (id) initWithService: (io_service_t) inServiceID
{
	if ((self = [super init]) != nil) {
		_serviceID = inServiceID;
		_locationID = [G15Device locationID: _serviceID];

		[self initHIDInterface];
		[self initDeviceInterface];

		if (_usbDevice == nil || _hidDevice == nil) {
			if (_usbDevice != nil) {
				(*_usbDevice)->Release(_usbDevice);
				_usbDevice = nil;
			}
			if (_hidDevice != nil) {
				(*_hidDevice)->Release(_hidDevice);
				_hidDevice = nil;
			}
			IOObjectRelease(_serviceID);
			self = nil;
		}
		else {
			[self completeStartup];
		}
	}
	
	return self;
}

- (IOReturn) initDeviceInterface
{
	CFMutableDictionaryRef  matchingDict;
	IOReturn				result;
	
	if ((matchingDict = IOServiceMatching(kIOUSBDeviceClassName)) == nil) {
		NSLog(@"IOServiceMatching failed File %s Line %d", __FILE__, __LINE__);
		result = kIOReturnError;
	}
	else {
		NSMutableDictionary		*dict = (NSMutableDictionary *)matchingDict;
		io_iterator_t			deviceIterator;
		io_object_t				usbDevice;
		
		[dict setValue: [NSNumber numberWithLong: kG15Product] forKey: [NSString stringWithCString: kUSBProductID]];
		[dict setValue: [NSNumber numberWithLong: kG15Vendor] forKey: [NSString stringWithCString: kUSBVendorID]];
		
		if ((result = IOServiceGetMatchingServices(kIOMasterPortDefault, matchingDict, &deviceIterator)) == kIOReturnSuccess) {
			IOCFPlugInInterface			**iodev = nil;
			SInt32						score;

			while ((usbDevice = IOIteratorNext(deviceIterator)) != 0 && !_usbDevice) {
				result = IOCreatePlugInInterfaceForService(usbDevice, kIOUSBDeviceUserClientTypeID,
																kIOCFPlugInInterfaceID, &iodev, &score);
				
				if (result == kIOReturnSuccess) {
					IOUSBDeviceInterface		**usbDeviceInterface;

					if ((result = (*iodev)->QueryInterface(iodev, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID), (LPVOID) &usbDeviceInterface)) == kIOReturnSuccess) {
						UInt32			deviceLocationID;
						
						if ((result = (*usbDeviceInterface)->GetLocationID(usbDeviceInterface, &deviceLocationID)) == kIOReturnSuccess) {
							if (deviceLocationID == _locationID)
								_usbDevice = usbDeviceInterface;
						}
					}
					
					
					if (_usbDevice == nil)
						IOObjectRelease(usbDevice);

					IODestroyPlugInInterface(iodev);
				}
			}
			
			IOObjectRelease(deviceIterator);
		}
	}
	
	if (result != kIOReturnSuccess)
		NSLog(@"initDeviceInterface failed: %08x", result);

	return result;
}

- (IOReturn) initHIDInterface
{
    IOCFPlugInInterface		**iodev = nil;
	IOReturn				result;
    SInt32					score;
	
	result = IOCreatePlugInInterfaceForService(_serviceID,
		kIOHIDDeviceUserClientTypeID,
		kIOCFPlugInInterfaceID, &iodev, &score);
	
	if (result == kIOReturnSuccess) {
		IOHIDDeviceInterface122		**hidDeviceInterface;

		if ((result = (*iodev)->QueryInterface(iodev, CFUUIDGetUUIDBytes(kIOHIDDeviceInterfaceID122), (LPVOID) &hidDeviceInterface)) == kIOReturnSuccess)
			_hidDevice = hidDeviceInterface;

		IODestroyPlugInInterface(iodev);
	}
	
	if (result != kIOReturnSuccess)
		NSLog(@"initHIDInterface failed: %08x", result);

	return result;
}

- (IOReturn) completeStartup
{
	CFRunLoopSourceRef		eventSource;
	IOReturn				result = kIOReturnError;	// assume failure (pessimist!)
 	
	if ((result = (*_hidDevice)->open(_hidDevice, 0)) == kIOReturnSuccess) {
		if ((result = (*_hidDevice)->createAsyncEventSource(_hidDevice, &eventSource)) != kIOReturnSuccess) {
			NSLog(@"completeStartup - createAsyncEventSource failed: %08x", result);
		}
		else {
			CFRunLoopAddSource(CFRunLoopGetCurrent(), eventSource, kCFRunLoopDefaultMode);
			result = (*_hidDevice)->setInterruptReportHandlerCallback(_hidDevice, _buffer, kG15ReportBufferSize, (IOHIDReportCallbackFunction)G15HIDReportCallbackFunction, self, nil);
		}
	}
	
	if (result != kIOReturnSuccess)
		NSLog(@"completeStartup failed: %08x", result);
	
	return result;
}

- (void) dealloc
{
	[self shutdownDevice];
	IOObjectRelease(_serviceID);
	
	[super dealloc];
}

- (void) shutdownDevice
{
	if (_hidDevice != nil) {
		(*_hidDevice)->close(_hidDevice);
		(*_hidDevice)->Release(_hidDevice);
		_hidDevice = nil;
	}
	if (_usbDevice != nil) {
		(*_usbDevice)->Release(_usbDevice);
		_usbDevice = nil;
	}
}

// byte 8 - bit 7 
// byte 2 - bit 7 
// byte 3 - bit 7 
// byte 4 - bit 7 
// byte 5 - bit 7 

- (void) processReadData: (UInt32) inBufferSize
{
	BOOL	key;
	
	key = (_buffer[8] & 0x80) ? YES : NO;
	if (key && !_clearKey)
		[[G15Display defaultDisplay] performClear];
	_clearKey = key;
	
	key = (_buffer[2] & 0x80) ? YES : NO;
	if (key && !_homeKey)
		[[G15Display defaultDisplay] performHome];
	_homeKey = key;
	
	key = (_buffer[3] & 0x80) ? YES : NO;
	if (key && !_scrollUpKey)
		[[G15Display defaultDisplay] performScrollUp];
	_scrollUpKey = key;
	
	key = (_buffer[4] & 0x80) ? YES : NO;
	if (key && !_scrollDownKey)
		[[G15Display defaultDisplay] performScrollDown];
	_scrollDownKey = key;
	
	key = (_buffer[5] & 0x80) ? YES : NO;
	if (key && !_endKey)
		[[G15Display defaultDisplay] performEnd];
	_endKey = key;
}

- (io_service_t) serviceID
{
	return _serviceID;
}

- (void) sendLCDData: (SLCDRec *) inData
{
	IOReturn			result;

	result = (*_hidDevice)->setReport(_hidDevice, kIOHIDReportTypeOutput, 0, inData, sizeof(SLCDRec), 1000, nil, nil, nil);
}

@end
